import validation
from invirt.common import CodeError, InvalidInput
import random
import sys
import time
import re
import cache_acls
import yaml

from invirt.config import structs as config
from invirt.database import Machine, Disk, Type, NIC, CDROM, session, meta
from invirt.remctl import remctl as gen_remctl

# ... and stolen from xend/uuid.py
def randomUUID():
    """Generate a random UUID."""

    return [ random.randint(0, 255) for _ in range(0, 16) ]

def uuidToString(u):
    """Turn a numeric UUID to a hyphen-seperated one."""
    return "-".join(["%02x" * 4, "%02x" * 2, "%02x" * 2, "%02x" * 2,
                     "%02x" * 6]) % tuple(u)
# end stolen code

def remctl(*args, **kwargs):
    return gen_remctl(config.remote.hostname,
                      principal='daemon/'+config.web.hostname,
                      *args, **kwargs)

def lvcreate(machine, disk):
    """Create a single disk for a machine"""
    remctl('web', 'lvcreate', machine.name,
           disk.guest_device_name, str(disk.size))
    
def makeDisks(machine):
    """Update the lvm partitions to add a disk."""
    for disk in machine.disks:
        lvcreate(machine, disk)

def getswap(disksize, memsize):
    """Returns the recommended swap partition size."""
    return int(min(disksize / 4, memsize * 1.5))

def lvinstall(machine, autoinstall):
    disksize = machine.disks[0].size
    memsize = machine.memory
    swapsize = getswap(disksize, memsize)
    imagesize = disksize - swapsize

    installer_options = ['dist=%s' % autoinstall.distribution,
                         'mirror=%s' % autoinstall.mirror,
                         'arch=%s' % autoinstall.arch,
                         'imagesize=%s' % imagesize]
    if autoinstall.preseed:
        installer_options += ['preseed=http://'+config.web.hostname+'/static/preseed/'+autoinstall.autoinstall_id+'.preseed']

    remctl('control', machine.name, 'install',
           *installer_options)

def lvcopy(machine_orig_name, machine, rootpw):
    """Copy a golden image onto a machine's disk"""
    remctl('web', 'lvcopy', machine_orig_name, machine.name, rootpw)

def bootMachine(machine, cdtype):
    """Boot a machine with a given boot CD.

    If cdtype is None, give no boot cd.  Otherwise, it is the string
    id of the CD (e.g. 'gutsy_i386')
    """
    if cdtype is not None:
        out, err = remctl('control', machine.name, 'create', 
                          cdtype, err=True)
    else:
        out, err = remctl('control', machine.name, 'create',
                          err=True)
    if 'already running' in err:
        raise InvalidInput('action', 'create',
                           'VM %s is already on' % machine.name)
    elif 'I need' in err and 'but dom0_min_mem is' in err:
        raise InvalidInput('action', 'create',
                           "We're really sorry, but our servers don't have enough capacity to create your VM right now. Try creating a VM with less RAM, or shutting down another VM of yours. Feel free to ask %s if you would like to know when we plan to have more resources." % (config.contact))
    elif ('Booting VMs is temporarily disabled for maintenance, sorry' in err or
          'LVM operations are temporarily disabled for maintenance, sorry' in err):
        raise InvalidInput('action', 'create',
                           err)
    elif "Boot loader didn't return any data!" in err:
        raise InvalidInput('action', 'create',
                           "The ParaVM bootloader was unable to find an operating system to boot. Do you have GRUB configured correctly?")
    elif 'xc_dom_find_loader: no loader found' in err:
        raise InvalidInput('action', 'create',
                           "The ParaVM bootloader was unable to boot the kernel you have configured. Are you sure this kernel is capable of running as a Xen ParaVM guest?")
    elif err:
        raise CodeError('"%s" on "control %s create %s' 
                        % (err, machine.name, cdtype))

def createVm(username, state, owner, contact, name, description, memory, disksize, machine_type, cdrom, autoinstall):
    """Create a VM and put it in the database"""
    # put stuff in the table
    session.begin()
    try:
        validation.Validate(username, state, name=name, description=description, owner=owner, memory=memory, disksize=disksize/1024.)
        machine = Machine()
        machine.name = name
        machine.description = description
        machine.memory = memory
        machine.owner = owner
        machine.administrator = None
        machine.contact = contact
        machine.uuid = uuidToString(randomUUID())
        machine.boot_off_cd = True
        machine.type = machine_type
        session.add(machine)
        disk = Disk(machine=machine,
                    guest_device_name='hda', size=disksize)
        nic = NIC.query.filter_by(machine_id=None).filter_by(reusable=True).first()
        if not nic: #No IPs left!
            raise CodeError("No IP addresses left!  "
                            "Contact %s." % config.contact)
        nic.machine = machine
        nic.hostname = name
        session.add(nic)
        session.add(disk)
        cache_acls.refreshMachine(machine)
        makeDisks(machine)
        session.commit()
    except:
        session.rollback()
        raise
    try:
        if autoinstall:
            lvinstall(machine, autoinstall)
        else:
            # tell it to boot with cdrom
            bootMachine(machine, cdrom)
    except CodeError, e:
        deleteVM(machine)
        raise
    return machine

def getList():
    """Return a dictionary mapping machine names to dicts."""
    value_string = remctl('web', 'listvms')
    value_dict = yaml.load(value_string, yaml.CSafeLoader)
    return value_dict

def parseStatus(s):
    """Parse a status string into nested tuples of strings.

    s = output of xm list --long <machine_name>
    """
    values = re.split('([()])', s)
    stack = [[]]
    for v in values[2:-2]: #remove initial and final '()'
        if not v:
            continue
        v = v.strip()
        if v == '(':
            stack.append([])
        elif v == ')':
            if len(stack[-1]) == 1:
                stack[-1].append('')
            stack[-2].append(stack[-1])
            stack.pop()
        else:
            if not v:
                continue
            stack[-1].extend(v.split())
    return stack[-1]

def statusInfo(machine):
    """Return the status list for a given machine.

    Gets and parses xm list --long
    """
    value_string, err_string = remctl('control', machine.name, 'list-long', 
                                      err=True)
    if 'Unknown command' in err_string:
        raise CodeError("ERROR in remctl list-long %s is not registered" % 
                        (machine.name,))
    elif 'is not on' in err_string:
        return None
    elif err_string:
        raise CodeError("ERROR in remctl list-long %s:  %s" % 
                        (machine.name, err_string))
    status = parseStatus(value_string)
    return status

def listHost(machine):
    """Return the host a machine is running on"""
    out, err = remctl('control', machine.name, 'listhost', err=True)
    if err:
        return None
    return out.strip()

def vnctoken(machine):
    """Return a time-stamped VNC token"""
    out, err = remctl('control', machine.name, 'vnctoken', err=True)
    if err:
        return None
    return out.strip()

def deleteVM(machine):
    """Delete a VM."""
    remctl('control', machine.name, 'destroy', err=True)
    session.begin()
    delete_disk_pairs = [(machine.name, d.guest_device_name) 
                         for d in machine.disks]
    try:
        for mname, dname in delete_disk_pairs:
            remctl('web', 'lvremove', mname, dname)
        for nic in machine.nics:
            nic.machine_id = None
            nic.hostname = None
            session.add(nic)
        for disk in machine.disks:
            session.delete(disk)
        session.delete(machine)
        session.commit()
    except:
        session.rollback()
        raise

def commandResult(username, state, command_name, machine_id, fields):
    start_time = 0
    result = None
    machine = validation.Validate(username, state, machine_id=machine_id).machine
    action = command_name
    cdrom = fields.get('cdrom') or None
    if cdrom is not None and not CDROM.query.filter_by(cdrom_id=cdrom).one():
        raise CodeError("Invalid cdrom type '%s'" % cdrom)    
    if action not in "reboot create destroy shutdown delete renumber".split(" "):
        raise CodeError("Invalid action '%s'" % action)
    if action == 'reboot':
        if cdrom is not None:
            out, err = remctl('control', machine.name, 'reboot', cdrom,
                              err=True)
        else:
            out, err = remctl('control', machine.name, 'reboot',
                              err=True)
        if err:
            if re.match("machine '.*' is not on", err):
                raise InvalidInput("action", "reboot", 
                                   "Machine is not on")
            else:
                print >> sys.stderr, 'Error on reboot:'
                print >> sys.stderr, err
                raise CodeError('ERROR on remctl')
                
    elif action == 'create':
        if validation.maxMemory(username, state, machine) < machine.memory:
            raise InvalidInput('action', 'Power on',
                               "You don't have enough free RAM quota "
                               "to turn on this machine.")
        bootMachine(machine, cdrom)
    elif action == 'destroy':
        out, err = remctl('control', machine.name, 'destroy', err=True)
        if err:
            if re.match("machine '.*' is not on", err):
                raise InvalidInput("action", "Power off", 
                                   "Machine is not on.")
            else:
                print >> sys.stderr, 'Error on power off:'
                print >> sys.stderr, err
                raise CodeError('ERROR on remctl')
    elif action == 'shutdown':
        out, err = remctl('control', machine.name, 'shutdown', err=True)
        if err:
            if re.match("machine '.*' is not on", err):
                raise InvalidInput("action", "Shutdown", 
                                   "Machine is not on.")
            else:
                print >> sys.stderr, 'Error on Shutdown:'
                print >> sys.stderr, err
                raise CodeError('ERROR on remctl')
    elif action == 'delete':
        deleteVM(machine)
    elif action == 'renumber':
        result = remctl('control', machine.name, 'renumber')

    d = dict(user=username,
             command=action,
             machine=machine)
    if result:
        d['result'] = result
    return d

def resizeDisk(machine_name, disk_name, new_size):
    remctl("web", "lvresize", machine_name, disk_name, new_size)

def renameMachine(machine, old_name, new_name):
    for disk in machine.disks:
        remctl("web", "lvrename", old_name, 
               disk.guest_device_name, new_name)
    
